import numpy as np
import pytest
from fixtures import three_storage_model

from pywr.core import Model, Input, Output, Link, Storage, AggregatedStorage
from pywr.parameters.control_curves import ControlCurveParameter

def test_aggregated_storage(three_storage_model):
    """ Test `AggregatedStorage` correct sums multiple `Storage`. """
    m = three_storage_model

    agg_stg = m.nodes['Total Storage']
    stgs = [m.nodes['Storage {}'.format(num)] for num in range(3)]
    # Check initial volume is aggregated correctly
    np.testing.assert_allclose(agg_stg.initial_volume, sum(s.initial_volume for s in stgs))

    m.setup()
    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))


def test_aggregated_storage_initial_pc(three_storage_model):
    """Test `AggregatedStorage` correct sums multiple `Storage` with proportional initial volumes."""
    m = three_storage_model

    agg_stg = m.nodes['Total Storage']
    stgs = [m.nodes['Storage {}'.format(num)] for num in range(3)]
    for stg in stgs:
        stg.initial_volume = None
        stg.initial_volume_pc = 0.9
    # Check initial volume is aggregated correctly
    np.testing.assert_allclose(agg_stg.initial_volume, sum(s.initial_volume_pc * s.max_volume for s in stgs))

    m.setup()
    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))


def test_aggregated_storage_scenarios(three_storage_model):
    """ Test `AggregatedScenario` correctly sums over scenarios. """
    from pywr.core import Scenario
    from pywr.parameters import ConstantScenarioParameter
    m = three_storage_model
    sc = Scenario(m, 'A', size=5)

    agg_stg = m.nodes['Total Storage']
    stgs = [m.nodes['Storage {}'.format(num)] for num in range(3)]
    # Setup scenario
    inpt1 = m.nodes['Input 1'].max_flow = ConstantScenarioParameter(m, sc, range(sc.size))

    m.setup()

    # Test initial volume
    np.testing.assert_allclose(agg_stg.volume, sum(s.initial_volume for s in stgs))
    current_pc = sum(s.initial_volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_stg.volume, sum(s.volume for s in stgs))
    current_pc = sum(s.volume for s in stgs) / (sum(s.max_volume for s in stgs))
    np.testing.assert_allclose(agg_stg.current_pc, current_pc)
    np.testing.assert_allclose(agg_stg.flow, sum(s.flow for s in stgs))


def test_aggregated_storage_connectivity(three_storage_model):
    """ Test `AggregatedStorage` has the correct attributes. """
    m = three_storage_model
    agg_stg = m.nodes['Total Storage']
    inpt = m.nodes['Input 1']

    with pytest.raises(AttributeError):
        agg_stg.connect()

    with pytest.raises(TypeError):
        # If not `Connectable` the node won't have a iter_slots attribute
        inpt.connect(agg_stg)


def test_aggregated_storage_attributes(three_storage_model):
    """ Test `AggregatedStorage` has the correct attributes. """
    m = three_storage_model
    agg_stg = m.nodes['Total Storage']

    with pytest.raises(AttributeError):
        agg_stg.get_max_volume()

    with pytest.raises(AttributeError):
        agg_stg.get_min_volume()


def test_aggregated_storage_control_curve(three_storage_model):
    """Test using a control curve based on an aggregate storage, rather than
    a single storage.
    """
    model = three_storage_model

    # create a new supply node
    inpt = Input(model, "Input 3", cost=-1000)
    inpt.connect(model.nodes["Output 0"])
    inpt.connect(model.nodes["Output 1"])
    inpt.connect(model.nodes["Output 2"])

    # limit the flow of the new node using a control curve on the aggregate storage
    curves = [0.5] # 50%
    values = [0, 5]
    inpt.max_flow = ControlCurveParameter(model, model.nodes["Total Storage"], curves, values)

    # initial storage is > 50% so flow == 0
    model.step()
    np.testing.assert_allclose(inpt.flow, 0.0)

    # set initial storage to < 50%
    storages = [node for node in model.nodes if isinstance(node, Storage)]
    for node, value in zip(storages, [0.6, 0.1, 0.1]):
        if isinstance(node, Storage):
            node.initial_volume = node.max_volume * value

    # now below the control curve, so flow is allowed
    model.reset()
    model.step()
    np.testing.assert_allclose(inpt.flow, 5.0)


def test_aggregated_node(three_storage_model):
    """ Test `AggregatedNode` correct sums multiple `Node`. """
    m = three_storage_model

    agg_otpt = m.nodes['Total Output']
    otpts = [m.nodes['Output {}'.format(num)] for num in range(3)]

    m.setup()
    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_otpt.flow, sum(o.flow for o in otpts))

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_otpt.flow, sum(o.flow for o in otpts))


def test_aggregated_node_scenarios(three_storage_model):
    """ Test `AggregatedScenario` correctly sums over scenarios. """
    from pywr.core import Scenario
    from pywr.parameters import ConstantScenarioParameter
    m = three_storage_model
    sc = Scenario(m, 'A', size=5)

    agg_otpt = m.nodes['Total Output']
    otpts = [m.nodes['Output {}'.format(num)] for num in range(3)]
    # Setup scenario
    inpt1 = m.nodes['Input 1'].max_flow = ConstantScenarioParameter(m, sc, range(sc.size))

    m.setup()
    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_otpt.flow, sum(o.flow for o in otpts))

    m.step()

    # Finally check volume is summed correctly
    np.testing.assert_allclose(agg_otpt.flow, sum(o.flow for o in otpts))


def test_aggregated_node_connectivity(three_storage_model):
    """ Test `AggregatedStorage` has the correct attributes. """
    m = three_storage_model
    agg_otpt = m.nodes['Total Output']
    inpt = m.nodes['Input 1']

    with pytest.raises(AttributeError):
        agg_otpt.connect()

    with pytest.raises(TypeError):
        # If not `Connectable` the node won't have a iter_slots attribute
        inpt.connect(agg_otpt)


def test_aggregated_node_attributes(three_storage_model):
    """ Test `AggregatedStorage` has the correct attributes. """
    m = three_storage_model
    agg_otpt = m.nodes['Total Output']

    with pytest.raises(AttributeError):
        agg_otpt.get_max_volume()

    with pytest.raises(AttributeError):
        agg_otpt.get_min_volume()




import numpy as np
import matplotlib.pyplot as plt
import pandas

baseline = 50.0

profile = np.array([0.9, 0.9, 0.9, 0.9, 1.2, 1.2, 1.2, 1.2, 0.9, 0.9, 0.9, 0.9])

L0 = np.array([1.0]*12)
L1 = np.array([0.95, 0.95, 0.95, 0.95, 0.90, 0.90, 0.90, 0.90, 0.95, 0.95, 0.95, 0.95])
L2 = np.array([0.8, 0.8, 0.8, 0.8, 0.75, 0.75, 0.75, 0.75, 0.8, 0.8, 0.8, 0.8])

dates = pandas.date_range("2015-01-01", "2015-12-31")

L0_values = np.array([baseline*profile[d.month-1]*L0[d.month-1] for d in dates])
L1_values = np.array([baseline*profile[d.month-1]*L1[d.month-1] for d in dates])
L2_values = np.array([baseline*profile[d.month-1]*L2[d.month-1] for d in dates])

fig = plt.figure(figsize=(8, 5), dpi=300)

plt.plot(np.arange(0, len(dates)), L0_values, linewidth=2)
plt.plot(np.arange(0, len(dates)), L1_values, linewidth=2)
plt.plot(np.arange(0, len(dates)), L2_values, linewidth=2)

plt.xlabel("Day of year")
plt.ylabel("Demand [Ml/d]")

plt.grid(True)
plt.ylim([0.0, 70])
plt.xlim(0, 365)
plt.legend(["Level 0", "Level 1", "Level 2"], loc="lower right")
plt.tight_layout()
plt.show()



import json
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns


if __name__ == '__main__':

    store = pd.HDFStore('two_reservoir_moea.h5')

    fig, ax = plt.subplots(figsize=(6, 6))
    colors = sns.color_palette()

    for color, key in zip(colors, store.keys()):
        df = store[key]
        label = key.replace('/', '').replace('_', ' ').capitalize()
        df.plot.scatter(ax=ax, x='deficit', y='transferred', label=label, color=color)

    ax.set_xlabel('Total deficit [$Mm^3$]')
    ax.set_ylabel('Total transfer [$Mm^3$]')
    ax.grid()
    fig.savefig('two_reservoir_pareto.png')

    plt.show()



from ._recorders import NumpyArrayNodeRecorder


cpdef double hydropower_calculation(double flow, double water_elevation, double turbine_elevation, double efficiency,
                                    double flow_unit_conversion=1.0, double energy_unit_conversion=1e-6,
                                    double density=1000.0):
    """
    Calculate the total power produced using the hydropower equation.
    
   
    Parameters
    ----------
    flow : double 
        Flow rate of water through the turbine. Should be converted using `flow_unit_conversion` to 
        units of $m^3£ per day (not per second).
    water_elevation : double
        Elevation of water entering the turbine. The difference of this value with the `turbine_elevation` gives
        the working head of the turbine.
    turbine_elevation : double
        Elevation of the turbine itself. The difference between the `water_elevation` and this value gives
        the working head of the turbine.
    efficiency : double
        An efficiency scaling factor for the power output of the turbine.
    flow_unit_conversion : double (default=1.0)
        A factor used to transform the units of flow to be compatible with the equation here. This
        should convert flow to units of $m^3/day$
    energy_unit_conversion : double (default=1e-6)
        A factor used to transform the units of power. Defaults to 1e-6 to return $MJ$/day. 
    density : double (default=1000)
        Density of water in $kgm^{-3}$.
        
    Returns
    -------
    power : double
        Hydropower production rate in units of energy per day.
    
    Notes
    -----
    The hydropower calculation uses the following equation.
    
    .. math:: P = \rho * g * \deltaH * q
    
    The flow rate in should be converted to units of :math:`m^3` per day using the `flow_unit_conversion` parameter.    
    
    """
    cdef double head
    cdef double power
    cdef double q

    head = water_elevation - turbine_elevation
    if head < 0.0:
        head = 0.0

    # Convert flow to correct units (typically to m3/day)
    q = flow * flow_unit_conversion
    # Power
    power = density * q * 9.81 * head * efficiency

    return power * energy_unit_conversion


cdef class HydropowerRecorder(NumpyArrayNodeRecorder):
    """ Calculates the power production using the hydropower equation

    This recorder saves an array of the hydrpower production in each timestep. It can be converted to a dataframe
    after a model run has completed. It does not calculate total energy production.

    Parameters
    ----------

    water_elevation_parameter : Parameter instance (default=None)
        Elevation of water entering the turbine. The difference of this value with the `turbine_elevation` gives
        the working head of the turbine.
    turbine_elevation : double
        Elevation of the turbine itself. The difference between the `water_elevation` and this value gives
        the working head of the turbine.
    efficiency : float (default=1.0)
        The efficiency of the turbine.
    density : float (default=1000.0)
        The density of water.
    flow_unit_conversion : float (default=1.0)
        A factor used to transform the units of flow to be compatible with the equation here. This
        should convert flow to units of :math:`m^3/day`
    energy_unit_conversion : float (default=1e-6)
        A factor used to transform the units of total energy. Defaults to 1e-6 to return :math:`MJ`.

    Notes
    -----
    The hydropower calculation uses the following equation.

    .. math:: P = \\rho * g * \\delta H * q

    The flow rate in should be converted to units of :math:`m^3` per day using the `flow_unit_conversion` parameter.

    Head is calculated from the given `water_elevation_parameter` and `turbine_elevation` value. If water elevation
    is given then head is the difference in elevation between the water and the turbine. If water elevation parameter
    is `None` then the head is simply the turbine elevation.


    See Also
    --------
    TotalHydroEnergyRecorder
    pywr.parameters.HydropowerTargetParameter

    """
    def __init__(self, model, node, water_elevation_parameter=None, turbine_elevation=0.0, efficiency=1.0, density=1000,
                 flow_unit_conversion=1.0, energy_unit_conversion=1e-6, **kwargs):
        super(HydropowerRecorder, self).__init__(model, node, **kwargs)

        self.water_elevation_parameter = water_elevation_parameter
        self.turbine_elevation = turbine_elevation
        self.efficiency = efficiency
        self.density = density
        self.flow_unit_conversion = flow_unit_conversion
        self.energy_unit_conversion = energy_unit_conversion

    property water_elevation_parameter:
        def __get__(self):
            return self._water_elevation_parameter
        def __set__(self, parameter):
            if self._water_elevation_parameter:
                self.children.remove(self._water_elevation_parameter)
            self.children.add(parameter)
            self._water_elevation_parameter = parameter

    cpdef after(self):
        cdef int i
        cdef double q, head, power
        cdef Timestep ts = self.model.timestepper.current
        cdef ScenarioIndex scenario_index
        flow = self.node.flow

        for scenario_index in self.model.scenarios.combinations:

            if self._water_elevation_parameter is not None:
                head = self._water_elevation_parameter.get_value(scenario_index)
                if self.turbine_elevation is not None:
                    head -= self.turbine_elevation
            elif self.turbine_elevation is not None:
                head = self.turbine_elevation
            else:
                raise ValueError('One or both of storage_node or level must be set.')

            # -ve head is not valid
            head = max(head, 0.0)
            # Get the flow from the current node
            q = self._node._flow[scenario_index.global_id]
            power = hydropower_calculation(q, head, 0.0, self.efficiency, density=self.density,
                                             flow_unit_conversion=self.flow_unit_conversion,
                                             energy_unit_conversion=self.energy_unit_conversion)

            self._data[ts.index, scenario_index.global_id] = power

    @classmethod
    def load(cls, model, data):
        from pywr.parameters import load_parameter
        node = model.nodes[data.pop("node")]
        if "water_elevation_parameter" in data:
            water_elevation_parameter = load_parameter(model, data.pop("water_elevation_parameter"))
        else:
            water_elevation_parameter = None

        return cls(model, node, water_elevation_parameter=water_elevation_parameter, **data)
HydropowerRecorder.register()


cdef class TotalHydroEnergyRecorder(BaseConstantNodeRecorder):
    """ Calculates the total energy production using the hydropower equation from a model run.

    This recorder saves the total energy production in each scenario during a model run. It does not save a timeseries
    or power, but rather total energy.

    Parameters
    ----------

    water_elevation_parameter : Parameter instance (default=None)
        Elevation of water entering the turbine. The difference of this value with the `turbine_elevation` gives
        the working head of the turbine.
    turbine_elevation : double
        Elevation of the turbine itself. The difference between the `water_elevation` and this value gives
        the working head of the turbine.
    efficiency : float (default=1.0)
        The efficiency of the turbine.
    density : float (default=1000.0)
        The density of water.
    flow_unit_conversion : float (default=1.0)
        A factor used to transform the units of flow to be compatible with the equation here. This
        should convert flow to units of :math:`m^3/day`
    energy_unit_conversion : float (default=1e-6)
        A factor used to transform the units of total energy. Defaults to 1e-6 to return :math:`MJ`.

    Notes
    -----
    The hydropower calculation uses the following equation.

    .. math:: P = \\rho * g * \\delta H * q

    The flow rate in should be converted to units of :math:`m^3` per day using the `flow_unit_conversion` parameter.

    Head is calculated from the given `water_elevation_parameter` and `turbine_elevation` value. If water elevation
    is given then head is the difference in elevation between the water and the turbine. If water elevation parameter
    is `None` then the head is simply the turbine elevation.


    See Also
    --------
    HydropowerRecorder
    pywr.parameters.HydropowerTargetParameter

    """
    def __init__(self, model, node, water_elevation_parameter=None, turbine_elevation=0.0, efficiency=1.0, density=1000,
                 flow_unit_conversion=1.0, energy_unit_conversion=1e-6, **kwargs):
        super(TotalHydroEnergyRecorder, self).__init__(model, node, **kwargs)

        self.water_elevation_parameter = water_elevation_parameter
        self.turbine_elevation = turbine_elevation
        self.efficiency = efficiency
        self.density = density
        self.flow_unit_conversion = flow_unit_conversion
        self.energy_unit_conversion = energy_unit_conversion

    property water_elevation_parameter:
        def __get__(self):
            return self._water_elevation_parameter
        def __set__(self, parameter):
            if self._water_elevation_parameter:
                self.children.remove(self._water_elevation_parameter)
            self.children.add(parameter)
            self._water_elevation_parameter = parameter

    cpdef after(self):
        """ Calculate the  """
        cdef int i
        cdef double q, head, power
        cdef Timestep ts = self.model.timestepper.current
        cdef double days = ts.days
        cdef ScenarioIndex scenario_index
        flow = self.node.flow

        for scenario_index in self.model.scenarios.combinations:

            if self._water_elevation_parameter is not None:
                head = self._water_elevation_parameter.get_value(scenario_index)
                if self.turbine_elevation is not None:
                    head -= self.turbine_elevation
            elif self.turbine_elevation is not None:
                head = self.turbine_elevation
            else:
                raise ValueError('One or both of storage_node or level must be set.')

            # -ve head is not valid
            head = max(head, 0.0)
            # Get the flow from the current node
            q = self._node._flow[scenario_index.global_id]
            power = hydropower_calculation(q, head, 0.0, self.efficiency, density=self.density,
                                             flow_unit_conversion=self.flow_unit_conversion,
                                             energy_unit_conversion=self.energy_unit_conversion)

            self._values[scenario_index.global_id] += power * days

    @classmethod
    def load(cls, model, data):
        from pywr.parameters import load_parameter
        node = model.nodes[data.pop("node")]
        if "water_elevation_parameter" in data:
            water_elevation_parameter = load_parameter(model, data.pop("water_elevation_parameter"))
        else:
            water_elevation_parameter = None

        return cls(model, node, water_elevation_parameter=water_elevation_parameter, **data)
TotalHydroEnergyRecorder.register()



import numpy as np
import matplotlib.pyplot as plt
import pandas

profile =[0.9, 0.9, 0.9, 0.9, 1.2, 1.2, 1.2, 1.2, 0.9, 0.9, 0.9, 0.9]

dates = pandas.date_range("2015-01-01", "2015-12-31")
values = [profile[d.month-1] for d in dates]

fig = plt.figure(figsize=(8, 5))

plt.plot(np.arange(0, len(dates)), values, linewidth=2)

plt.xlabel("Day of year")
plt.ylabel("Demand factor")

plt.grid(True)
plt.ylim([0.0, 1.5])
plt.xlim(0, 365)
plt.tight_layout()
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import datetime
import pytest
import pandas
from numpy.testing import assert_allclose

import pywr.core
from pywr.model import Model, ModelStructureError, ModelResult
from pywr.nodes import Storage, Input, Output, Link
import pywr.solvers
import pywr.parameters.licenses
import pywr.domains.river
from pywr.recorders import assert_rec
from helpers import load_model

import pywr.parameters


def test_run_simple1():
    '''Test the most basic model possible'''
    # parse the JSON into a model
    model = load_model('simple1.json')

    # run the model
    t0 = model.timestepper.start.to_pydatetime()
    model.step()

    # check results
    demand1 = model.nodes['demand1']
    assert_allclose(demand1.flow, 10.0, atol=1e-7)
    # initially the timestepper returns the first time-step, so timestepper.current
    # does not change after the first 'step'.
    assert(model.timestepper.current.datetime - t0 == datetime.timedelta(0))
    # check the timestamp incremented
    model.step()
    assert(model.timestepper.current.datetime - t0 == datetime.timedelta(1))

def test_model_results():
    '''Test model results object'''
    import pywr

    model = load_model('simple1.json')
    res = model.run()
    assert (isinstance(res, ModelResult))
    assert (res.timesteps == 365)
    assert (res.version == pywr.__version__)
    assert res.solver_stats['number_of_cols']
    assert res.solver_stats['number_of_rows']
    assert res.solver_name == model.solver.name
    print(res)
    print(res._repr_html_())


@pytest.mark.parametrize("json_file", ['reservoir1.json', 'reservoir1_pc.json'])
def test_run_reservoir1(json_file):
    '''Test a reservoir with no refill

    Without an additional supply the reservoir should empty and cause a failure.
    '''
    model = load_model(json_file)
    demand1 = model.nodes['demand1']
    supply1 = model.nodes['supply1']
    for demand, stored in [(10.0, 25.0), (10.0, 15.0), (10.0, 5.0), (5.0, 0.0), (0.0, 0.0)]:
        result = model.step()
        assert_allclose(demand1.flow, demand, atol=1e-7)
        assert_allclose(supply1.volume, stored, atol=1e-7)


def test_run_reservoir2():
    '''Test a reservoir fed by a river abstraction

    The river abstraction should refill the reservoir, but not quickly enough
    to keep up with the demand.
    '''
    model = load_model('reservoir2.json')

    demand1 = model.nodes['demand1']
    supply1 = model.nodes['supply1']
    catchment = model.nodes['catchment1']
    assert(catchment.min_flow == 5)
    for demand, stored in [(15.0, 25.0), (15.0, 15.0), (15.0, 5.0), (10.0, 0.0), (5.0, 0.0)]:
        result = model.step()
        assert_allclose(demand1.flow[0], demand, atol=1e-7)
        assert_allclose(supply1.volume[0], stored, atol=1e-7)

def test_empty_storage_min_flow():

    model = Model()
    storage = Storage(model, "storage", initial_volume=100, max_volume=100, inputs=1, outputs=0)
    otpt = Output(model, "output", min_flow=75)
    storage.connect(otpt)
    model.check()
    model.step()
    with pytest.raises(RuntimeError):
        model.step()

def test_run_river1():
    '''Test a river abstraction with a simple catchment'''
    model = load_model('river1.json')

    result = model.step()
    demand1 = model.nodes['demand1']
    assert_allclose(demand1.flow, 5.0, atol=1e-7)


def test_run_river2():
    '''Test a river abstraction with two catchments, a confluence and a split'''
    model = load_model('river2.json')

    model.step()

    demand1 = model.nodes['demand1']
    assert_allclose(demand1.flow, 7.25, atol=1e-7)
    demand2 = model.nodes['demand2']
    assert_allclose(demand2.flow, 2.0, atol=1e-7)


# Contains an out of range date for pandas.to_datetime
@pytest.mark.parametrize("json_file", ['timeseries1.json', 'timeseries1_xlsx.json'])
def test_run_timeseries1(json_file):
    model = load_model(json_file)

    # check first day initalised
    assert(model.timestepper.start == datetime.datetime(2015, 1, 1))

    # check results
    demand1 = model.nodes['demand1']
    catchment1 = model.nodes['catchment1']
    for expected in (23.92, 22.14, 22.57, 24.97, 27.59):
        result = model.step()
        assert_allclose(catchment1.flow, expected, atol=1e-7)
        assert_allclose(demand1.flow, min(expected, 23.0), atol=1e-7)


# Contains an out of range date for pandas.to_datetime
@pytest.mark.parametrize("json_file", ['timeseries1_weekly.json', 'timeseries1_weekly_hdf.json'])
def test_run_timeseries1_weekly(json_file):
    model = load_model(json_file)

    # check first day initalised
    assert(model.timestepper.start == datetime.datetime(2015, 1, 1))

    # check results
    demand1 = model.nodes['demand1']
    catchment1 = model.nodes['catchment1']
    for expected in (23.92, 25.67, 28.24, 25.28, 21.84):
        result = model.step()
        assert_allclose(catchment1.flow, expected, atol=1e-7)
        assert_allclose(demand1.flow, min(expected, 23.0), atol=1e-7)


def test_run_cost1():
    model = load_model('cost1.json')

    supply1 = model.nodes['supply1']
    supply2 = model.nodes['supply2']
    demand1 = model.nodes['demand1']

    assert_allclose(supply1.get_cost(None), 1)
    assert_allclose(supply2.get_cost(None), 2)  # more expensive

    result = model.step()
    # check entire demand was supplied by supply1
    assert_allclose(supply1.flow, 10.0, atol=1e-7)
    assert_allclose(supply2.flow, 0.0, atol=1e-7)
    assert_allclose(demand1.flow, 10.0, atol=1e-7)

    # increase demand to more than supply1 can provide on it's own
    # and check that supply2 is used to pick up the slack
    demand1.max_flow = 20.0
    model.setup()
    result = model.step()
    assert_allclose(supply1.flow, 15.0, atol=1e-7)
    assert_allclose(supply2.flow, 5.0, atol=1e-7)
    assert_allclose(demand1.flow, 20.0, atol=1e-7)

    # supply as much as possible, even if it isn't enough
    demand1.max_flow = 40.0
    model.setup()
    result = model.step()
    assert_allclose(supply1.flow, 15.0, atol=1e-7)
    assert_allclose(supply2.flow, 15.0, atol=1e-7)
    assert_allclose(demand1.flow, 30.0, atol=1e-7)


def test_run_bottleneck():
    '''Test max flow constraint on intermediate nodes is upheld'''
    model = load_model('bottleneck.json')
    result = model.step()
    d1 = model.nodes['demand1']
    d2 = model.nodes['demand2']
    assert_allclose(d1.flow+d2.flow, 15.0, atol=1e-7)

@pytest.mark.skipif(Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
def test_run_discharge_upstream():
    '''Test river with inline discharge (upstream)

    In this instance the discharge is upstream of the abstraction, and so can
    be abstracted in the same way as the water from the catchment
    '''
    model = load_model('river_discharge1.json')
    model.step()
    demand = model.nodes['demand1']
    term = model.nodes['term1']
    assert_allclose(demand.flow, 8.0, atol=1e-7)
    assert_allclose(term.flow, 0.0, atol=1e-7)

@pytest.mark.skipif(Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
def test_run_discharge_downstream():
    '''Test river with inline discharge (downstream)

    In this instance the discharge is downstream of the abstraction, so the
    water shouldn't be available.
    '''
    model = load_model('river_discharge2.json')
    model.step()
    demand = model.nodes['demand1']
    term = model.nodes['term1']
    assert_allclose(demand.flow, 5.0, atol=1e-7)
    assert_allclose(term.flow, 3.0, atol=1e-7)

def test_new_storage():
    """Test new-style storage node with multiple inputs"""
    model = pywr.core.Model(
        start=pandas.to_datetime('1888-01-01'),
        end=pandas.to_datetime('1888-01-01'),
        timestep=datetime.timedelta(1)
    )

    supply1 = pywr.core.Input(model, 'supply1')

    splitter = pywr.core.Storage(model, 'splitter', outputs=1, inputs=2, max_volume=10, initial_volume=5)

    demand1 = pywr.core.Output(model, 'demand1')
    demand2 = pywr.core.Output(model, 'demand2')

    supply1.connect(splitter)

    splitter.connect(demand1, from_slot=0)
    splitter.connect(demand2, from_slot=1)

    supply1.max_flow = 45.0
    demand1.max_flow = 20
    demand2.max_flow = 40

    demand1.cost = -150
    demand2.cost = -100

    model.run()

    assert_allclose(supply1.flow, [45], atol=1e-7)
    assert_allclose(splitter.volume, [0], atol=1e-7)  # New volume is zero
    assert_allclose(demand1.flow, [20], atol=1e-7)
    assert_allclose(demand2.flow, [30], atol=1e-7)


def test_virtual_storage():
    """ Test the VirtualStorage node """

    model = pywr.core.Model()

    inpt = Input(model, "Input", max_flow=20)
    lnk = Link(model, "Link")
    inpt.connect(lnk)
    otpt = Output(model, "Output", max_flow=10, cost=-10.0)
    lnk.connect(otpt)

    vs = pywr.core.VirtualStorage(model, "Licence", [lnk], initial_volume=10.0, max_volume=10.0)

    model.setup()

    assert_allclose(vs.volume, [10], atol=1e-7)

    model.step()

    assert_allclose(otpt.flow, [10], atol=1e-7)
    assert_allclose(vs.volume, [0], atol=1e-7)

    model.step()

    assert_allclose(otpt.flow, [0], atol=1e-7)
    assert_allclose(vs.volume, [0], atol=1e-7)


def test_virtual_storage_duplicate_route():
    """ Test the VirtualStorage node """

    model = pywr.core.Model()

    inpt = Input(model, "Input", max_flow=20)
    lnk = Link(model, "Link")
    inpt.connect(lnk)
    otpt = Output(model, "Output", max_flow=10, cost=-10.0)
    lnk.connect(otpt)

    vs = pywr.core.VirtualStorage(model, "Licence", [lnk, otpt], factors=[0.5, 1.0], initial_volume=10.0, max_volume=10.0)

    model.setup()

    assert_allclose(vs.volume, [10], atol=1e-7)

    model.step()

    assert_allclose(otpt.flow, [10/1.5], atol=1e-7)
    assert_allclose(vs.volume, [0], atol=1e-7)

    model.step()

    assert_allclose(otpt.flow, [0], atol=1e-7)
    assert_allclose(vs.volume, [0], atol=1e-7)


class TestRollingVirtualStorage:

    @pytest.mark.parametrize('from_json', [True, False])
    def test_run(self, from_json):
        """Test RollingVirtualStorage node behaviour."""

        if from_json:
            model = load_model('rolling_virtual_storage.json')

            vs = model.nodes['Licence']
            otpt = model.nodes['Output']
        else:
            model = pywr.core.Model()

            inpt = Input(model, "Input", max_flow=20)
            lnk = Link(model, "Link")
            inpt.connect(lnk)
            otpt = Output(model, "Output", max_flow=15, cost=-10.0)
            lnk.connect(otpt)

            vs = pywr.core.RollingVirtualStorage(model, "Licence", [lnk], days=3, initial_volume=30.0, max_volume=30.0)

        model.setup()

        assert_allclose(vs.volume, [30], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [15], atol=1e-7)
        assert_allclose(vs.volume, [15], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [15], atol=1e-7)
        assert_allclose(vs.volume, [0], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [0], atol=1e-7)
        # End of third day the flow from the first day is return to the licence
        assert_allclose(vs.volume, [15], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [15], atol=1e-7)
        assert_allclose(vs.volume, [15], atol=1e-7)

    def test_run_weekly(self):
        """Test RollingVirtualStorage node behaviour with a weekly timestep."""

        model = pywr.core.Model(timestep='7D')

        inpt = Input(model, "Input", max_flow=20)
        lnk = Link(model, "Link")
        inpt.connect(lnk)
        otpt = Output(model, "Output", max_flow=10, cost=-10.0)
        lnk.connect(otpt)

        vs = pywr.core.RollingVirtualStorage(model, "Licence", [lnk], timesteps=3, initial_volume=100.0,
                                             max_volume=100.0)

        model.setup()

        assert_allclose(vs.volume, [100.0], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [10.0], atol=1e-7)
        assert_allclose(vs.volume, [30.0], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [30.0 / 7], atol=1e-7)
        assert_allclose(vs.volume, [0], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [0], atol=1e-7)
        # End of third day the flow from the first day is return to the licence
        assert_allclose(vs.volume, [70.0], atol=1e-7)

        model.step()

        assert_allclose(otpt.flow, [10.0], atol=1e-7)
        assert_allclose(vs.volume, [30.0], atol=1e-7)


def test_annual_virtual_storage():
    model = load_model('virtual_storage1.json')
    model.run()
    node = model.nodes["supply1"]
    rec = node.recorders[0]
    assert_allclose(rec.data[0], 10) # licence is not a constraint
    assert_allclose(rec.data[19], 10)
    assert_allclose(rec.data[20], 5) # licence is constraint
    assert_allclose(rec.data[21], 0) # licence is exhausted
    assert_allclose(rec.data[365], 10) # licence is refreshed


@pytest.mark.parametrize('reset_to_initial_volume', [None, False, True])
def test_annual_virtual_storage_reset_to_max_volume(reset_to_initial_volume):
    """Test that AnnualVirtualStorage resets to maximum volume. """
    model = load_model('virtual_storage1.json')
    licence1 = model.nodes['licence1']
    if reset_to_initial_volume is not None:
        licence1.reset_to_initial_volume = reset_to_initial_volume
    licence1.initial_volume = 100
    licence1.reset_month = 4

    model.setup()
    # After reset the current volume is always the initial volume
    assert_allclose(licence1.volume, [100.0])


    model.timestepper.start = '2015-04-01'
    model.reset()
    # After stepping over the reset day the volume should have been reset
    # before the solve to either initial volume or maximum volume.
    model.step()
    if reset_to_initial_volume:
        expected_volume = 90.0
    else:
        expected_volume = 195.0
    assert_allclose(licence1.volume, [expected_volume])


def test_annual_virtual_storage_with_dynamic_cost():
    model = load_model('virtual_storage2.json')
    model.run()
    node = model.nodes["supply1"]
    rec = node.recorders[0]

    assert_allclose(rec.data[0], 10)  # licence is not a constraint
    assert_allclose(rec.data[1], 5)  # now used slightly too much; switch to the other source
    assert_allclose(rec.data[2], 10)  # continue back and forth.
    assert_allclose(rec.data[3], 5)


class TestSeasonalVirtualStorage:
    def test_run(self):

        model = load_model('seasonal_virtual_storage.json')
        model.run()
        supply_df = model.recorders["supply1"].to_dataframe()
        licence_df = model.recorders["licence1"].to_dataframe()

        # License is not constraining flow as volumne remains
        assert_allclose(supply_df.loc["2015-01-01", :], 10)

        # License is depleted and constrains flow
        assert_allclose(supply_df.loc["2015-01-11", :], 0)
        assert_allclose(licence_df.loc["2015-01-11", :], 0)

        # License is depleted but does not constrain flow as it is not active
        assert_allclose(supply_df.loc["2015-02-01", :], 10)
        assert_allclose(licence_df.loc["2015-02-01", :], 0)

        # check same values for second year
        assert_allclose(supply_df.loc["2016-01-01", :], 10)
        assert_allclose(supply_df.loc["2016-01-11", :], 0)
        assert_allclose(licence_df.loc["2016-01-11", :], 0)
        assert_allclose(supply_df.loc["2016-02-01", :], 10)
        assert_allclose(licence_df.loc["2016-02-01", :], 0)

    def test_year_overlap(self):
        """test the the node works: end date < reset date < model start date. This
        means that the node's active period extends from one year to the next
        """

        model = load_model('seasonal_virtual_storage.json')

        vs = model.nodes["licence1"]
        vs.reset_day = 1
        vs.reset_month = 12
        vs.end_day = 31
        vs.end_month = 3
        model.timestepper.start = "2015-12-15"
        model.timestepper.end = "2016-12-02"

        model.run()
        supply_df = model.recorders["supply1"].to_dataframe()
        licence_df = model.recorders["licence1"].to_dataframe()

        # Start date is after reset data so there should be flow and volume should be reduced
        assert_allclose(supply_df.loc["2015-12-15", :], 10)
        assert_allclose(licence_df.loc["2015-12-15", :], 90)

        # Licence is depleted but remains active so flow is 0
        assert_allclose(supply_df.loc["2015-12-31", :], 0)
        assert_allclose(licence_df.loc["2015-12-31", :], 0)

        # License is turned off so flow is not constrained
        assert_allclose(supply_df.loc["2016-03-31", :], 10)

        # License is reset and made active
        assert_allclose(supply_df.loc["2016-12-01", :], 10)
        assert_allclose(licence_df.loc["2016-12-01", :], 90)

    def test_start_deactivated(self):
        """test the the node is not active when:  model start date < reset date < end date
        """

        model = load_model('seasonal_virtual_storage.json')

        vs = model.nodes["licence1"]
        vs.reset_day = 1
        vs.reset_month = 2
        vs.end_day = 1
        vs.end_month = 3
        model.timestepper.start = "2015-01-01"
        model.timestepper.end = "2015-04-01"

        model.run()
        supply_df = model.recorders["supply1"].to_dataframe()
        licence_df = model.recorders["licence1"].to_dataframe()

        # Start date before reset date and end date so there should be flow but no reduction in node storage
        assert_allclose(supply_df.loc["2015-01-15", :], 10)
        assert_allclose(licence_df.loc["2015-01-15", :], 100)

        # Licence active but depleted so flow is 0
        assert_allclose(supply_df.loc["2015-2-15", :], 0)
        assert_allclose(licence_df.loc["2015-2-15", :], 0)

        # License is turned off so flow is not constrained
        assert_allclose(supply_df.loc["2015-03-01", :], 10)


def test_storage_spill_compensation():
    """Test storage spill and compensation flows

    The upstream catchment has min_flow == max_flow, so it "pushes" water into
    the reservoir. The reservoir is already at it's maximum volume, so the
    water must go *somewhere*. The compensation flow has the most negative cost,
    so that is satisfied first. Once that is full, the demand is supplied.
    Finally, any surplus is forced into the spill despite the cost.

    Catchment -> Reservoir -> Demand
                         |--> Spill        --|
                         |--> Compensation --|
                                             |--> Terminator
    """
    model = pywr.core.Model()

    catchment = pywr.core.Input(model, name="Input", min_flow=10.0, max_flow=10.0, cost=1)
    reservoir = pywr.core.Storage(model, name="Storage", max_volume=100, initial_volume=100.0)
    spill = pywr.core.Link(model, name="Spill", cost=1.0)
    compensation = pywr.core.Link(model, name="Compensation", max_flow=3.0, cost=-999)
    terminator = pywr.core.Output(model, name="Terminator", cost=-1.0)
    demand = pywr.core.Output(model, name="Demand", max_flow=5.0, cost=-500)

    catchment.connect(reservoir)
    reservoir.connect(spill)
    reservoir.connect(compensation)
    reservoir.connect(demand)
    spill.connect(terminator)
    compensation.connect(terminator)

    model.check()
    model.run()
    assert_allclose(catchment.flow[0], 10.0, atol=1e-7)
    assert_allclose(demand.flow[0], 5.0, atol=1e-7)
    assert_allclose(compensation.flow[0], 3.0, atol=1e-7)
    assert_allclose(spill.flow[0], 2.0, atol=1e-7)
    assert_allclose(terminator.flow[0], (compensation.flow[0] + spill.flow[0]), atol=1e-7)


def test_reservoir_circle():
    """
    Issue #140. A model with a circular route, from a reservoir Input back
    around to it's own Output.

                 Demand
                    ^
                    |
                Reservoir <- Pumping
                    |           ^
                    v           |
              Compensation      |
                    |           |
                    v           |
    Catchment -> River 1 -> River 2 ----> MRFA -> Waste
                                    |              ^
                                    |---> MRFB ----|
    """
    model = Model()

    catchment = Input(model, "catchment", max_flow=500, min_flow=500)

    reservoir = Storage(model, "reservoir", max_volume=10000, initial_volume=5000)

    demand = Output(model, "demand", max_flow=50, cost=-100)
    pumping_station = Link(model, "pumping station", max_flow=100, cost=-10)
    river1 = Link(model, "river1")
    river2 = Link(model, "river2")
    compensation = Link(model, "compensation", cost=600)
    mrfA = Link(model, "mrfA", cost=-500, max_flow=50)
    mrfB = Link(model, "mrfB")
    waste = Output(model, "waste")

    catchment.connect(river1)
    river1.connect(river2)
    river2.connect(mrfA)
    river2.connect(mrfB)
    mrfA.connect(waste)
    mrfB.connect(waste)
    river2.connect(pumping_station)
    pumping_station.connect(reservoir)
    reservoir.connect(compensation)
    compensation.connect(river1)
    reservoir.connect(demand)

    model.check()
    model.setup()

    # not limited by mrf, pump capacity is constraint
    model.step()
    assert_allclose(catchment.flow, 500)
    assert_allclose(waste.flow, 400)
    assert_allclose(compensation.flow, 0)
    assert_allclose(pumping_station.flow, 100)
    assert_allclose(demand.flow, 50)

    # limited by mrf
    catchment.min_flow = catchment.max_flow = 100
    model.setup()
    model.step()
    assert_allclose(waste.flow, 50)
    assert_allclose(compensation.flow, 0)
    assert_allclose(pumping_station.flow, 50)
    assert_allclose(demand.flow, 50)

    # reservoir can support mrf, but doesn't need to
    compensation.cost = 200
    model.setup()
    model.step()
    assert_allclose(waste.flow, 50)
    assert_allclose(compensation.flow, 0)
    assert_allclose(pumping_station.flow, 50)
    assert_allclose(demand.flow, 50)

    # reservoir supporting mrf
    catchment.min_flow = catchment.max_flow = 0
    model.setup()
    model.step()
    assert_allclose(waste.flow, 50)
    assert_allclose(compensation.flow, 50)
    assert_allclose(pumping_station.flow, 0)
    assert_allclose(demand.flow, 50)


def test_breaklink_node():
    model = load_model('breaklink.json')
    supply = model.nodes["A"]
    transfer = model.nodes["B"]
    demand = model.nodes["C"]
    model.check()
    model.run()
    assert_allclose(supply.flow, 20)
    assert_allclose(transfer.flow, 20)
    assert_allclose(demand.flow, 20)
    assert_allclose(transfer.storage.volume, 0)


@pytest.mark.parametrize('loss_factor', [None, 0.2, 0.99, 1.0, 0.0])
def test_loss_link_node(loss_factor):
    """Test LossLink node"""
    model = load_model('loss_link.json')

    supply1 = model.nodes["supply1"]
    link1 = model.nodes["link1"]
    demand1 = model.nodes["demand1"]

    if loss_factor is not None:
        link1.loss_factor = loss_factor

    model.check()
    model.run()

    if loss_factor is None:
        expected_supply = 12
        expected_demand = 10
    elif loss_factor == 1.0:
        # 100% loss means no flow can be provided.
        expected_supply = 0.0
        expected_demand = 0.0
    else:
        expected_supply = 10 * (1 + loss_factor)
        expected_demand = 10

    # Supply must provide 20% more flow because of the loss in link1
    assert_allclose(supply1.flow, expected_supply)
    # link1 records the net flow after losses
    assert_allclose(link1.flow, expected_demand)
    assert_allclose(demand1.flow, expected_demand)


def test_reservoir_surface_area():
    from pywr.parameters import InterpolatedVolumeParameter
    model = load_model('reservoir_evaporation.json')
    model.timestepper.start = "1920-01-01"
    model.timestepper.end = "1920-01-02"
    res = model.run()
    assert hasattr(Storage, "area")
    assert isinstance(model.nodes["reservoir1"].area, InterpolatedVolumeParameter)
    assert_allclose(model.nodes["evaporation"].flow, 2.46875)


def test_reservoir_surface_area_without_area_property():
    """ Temporary test while the above test is not working. """
    model = load_model('reservoir_evaporation_without_area_property.json')
    model.timestepper.start = "1920-01-01"
    model.timestepper.end = "1920-01-02"
    res = model.run()
    assert_allclose(model.nodes["evaporation"].flow, 2.46875)

def test_reservoir_surface_area_external_data():
    """ Testing the InterpolatedVolume parameter with external data as
    inputs of Volume and Values, in this case a .csv file. """
    model = load_model('reservoir_evaporation_areafromfile.json')
    model.timestepper.start = "1920-01-01"
    model.timestepper.end = "1920-01-02"
    res = model.run()
    assert_allclose(model.nodes["evaporation"].flow, 2.46875)



def test_run_empty():
    # empty model should raise an exception if run
    model = Model()
    with pytest.raises(ModelStructureError):
        model.run()


def test_run():
    model = load_model('simple1.json')

    # run model from start to finish
    result = model.run()
    assert(result.timestep.index == 364)

    # reset model and run again
    result = model.run()
    assert(result.timestep.index == 364)

    # run remaining timesteps
    model.reset(start=pandas.to_datetime('2015-12-01'))
    result = model.run()
    assert(result.timestep.index == 364)


def test_reset_prev_flow():
    """Test resetting the prev_flow attribte of a node."""
    model = load_model('simple1.json')
    demand1 = model.nodes['demand1']
    model.run()
    assert_allclose(demand1.flow, 10.0, atol=1e-7)
    assert_allclose(demand1.prev_flow, 10.0, atol=1e-7)
    # Reset and check flow attributes are zeroed
    model.reset()
    assert_allclose(demand1.flow, 0.0, atol=1e-7)
    assert_allclose(demand1.prev_flow, 0.0, atol=1e-7)
    # Run again
    model.run()
    assert_allclose(demand1.flow, 10.0, atol=1e-7)
    assert_allclose(demand1.prev_flow, 10.0, atol=1e-7)


def test_run_monthly():
    model = load_model('simple1_monthly.json')

    result = model.run()
    assert result.timestep.index == 11

    result = model.run()
    assert result.timestep.index == 11


def test_select_solver():
    """Test specifying the solver in JSON"""
    solver_names = [solver.name for solver in pywr.solvers.solver_registry]
    for solver_name in solver_names:
        data = '''{"metadata": {"minimum_version": "0.1"}, "nodes": {}, "edges": {}, "timestepper": {"start": "1990-01-01","end": "1999-12-31","timestep": 1}, "solver": {"name": "%s"}}''' % solver_name
        model = load_model(data=data)
        assert(model.solver.name.lower() == solver_name)

def test_solver_unrecognised():
    '''Test specifying an unrecognised solver JSON'''
    solver_name = 'foobar'
    data = '''{"metadata": {"minimum_version": "0.1"}, "nodes": {}, "edges": {}, "timestepper": {"start": "1990-01-01","end": "1999-12-31","timestep": 1}, "solver": {"name": "%s"}}''' % solver_name
    with pytest.raises(KeyError):
        model = load_model(data=data)

@pytest.mark.skipif(Model().solver.name != "glpk", reason="only valid for glpk")
@pytest.mark.parametrize("use_presolve", ["true", "false"])
def test_select_glpk_presolve(use_presolve):
    """Test specifying the solver in JSON"""
    solver_names = ["glpk"]
    for solver_name in solver_names:
        data = '''{"metadata": {"minimum_version": "0.1"}, "nodes": {}, "edges": {}, "timestepper": {"start": "1990-01-01","end": "1999-12-31","timestep": 1}, "solver": {"name": "%s", "use_presolve": %s}}''' % (solver_name, use_presolve)
        model = load_model(data=data)
        assert(model.solver.name.lower() == solver_name)
        assert(model.solver._cy_solver.use_presolve == (use_presolve == "true"))



import os
import numpy as np
import pytest
from platypus import NSGAII
from pywr.utils.bisect import BisectionSearchModel
from pywr.optimisation.platypus import PlatypusWrapper
from pywr.optimisation import clear_global_model_cache
from helpers import load_model

TEST_FOLDER = os.path.dirname(__file__)


def test_simple_bisection():
    """Test basic bisection search"""

    model = load_model("simple1_bisect.json", model_klass=BisectionSearchModel)
    assert isinstance(model, BisectionSearchModel)

    # Perform bisection search
    model.run()

    np.testing.assert_allclose(model.parameters["demand"].get_double_variables(), [17.5])


def test_simple_infeasible_bisection():
    """Test infeasible bisection search"""

    model = load_model("simple1_infeasible_bisect.json", model_klass=BisectionSearchModel)
    assert isinstance(model, BisectionSearchModel)

    # Perform bisection search
    with pytest.raises(ValueError):
        model.run()


@pytest.fixture()
def simple_bisection_problem():
    filename = os.path.join(TEST_FOLDER, 'models', 'simple1_bisect.json')
    yield PlatypusWrapper(filename, model_klass=BisectionSearchModel)
    # Clean up the
    clear_global_model_cache()
    # We force deallocation the cache here to prevent problems using process pools
    # with pytest.
    import gc
    gc.collect()


def test_platypus_init(simple_bisection_problem):
    """ Test the initialisation of the platypus problem. """
    p = simple_bisection_problem

    assert p.problem.nvars == 1
    assert p.problem.nobjs == 1
    assert p.problem.nconstrs == 1
    # Check the correct model class is created by the wrapper.
    assert isinstance(p.model, BisectionSearchModel)


def test_platypus_nsgaii_step(simple_bisection_problem):
    """ Undertake a single step of the NSGAII algorithm with a small population. """
    algorithm = NSGAII(simple_bisection_problem.problem, population_size=10)
    algorithm.step()



"""
This module contains a Solver baseclass and several implemented subclasses.

Solvers are used to with pywr.core.Model classes to solve the network
allocation problem every time step.

Currently there are only linear programme based solvers using,
    - GLPK
    - LPSolve55
"""
import os
solver_registry = []


def _parse_env_kwarg(kwargs, keyword, env_name, env_type):
    """Insert a keyword argument from an environment."""
    if keyword not in kwargs:
        # Keyword not given by user; see if an environment variable is defined with one
        env_value = os.environ.get(env_name, None)
        if env_value is not None:
            if env_type is bool:
                kwargs[keyword] = env_value == '1' or env_value.lower() == 'true'
            elif env_type is str:
                kwargs[keyword] = env_value
            elif env_type is float:
                kwargs[keyword] = float(env_value)
            elif env_type is int:
                kwargs[keyword] = int(env_value)
            else:
                raise NotImplementedError(f'Parsing environment variables of type {env_type} is not supported.')
    return kwargs


class Solver(object):
    """Solver base class from which all solvers should inherit"""
    name = 'default'

    def __init__(self, *args, **kwargs):
        pass

    def setup(self, model):
        raise NotImplementedError('Solver should be subclassed to provide setup()')

    def solve(self, model):
        raise NotImplementedError('Solver should be subclassed to provide solve()')

    def reset(self):
        raise NotImplementedError('Solver should be subclassed to provide reset()')

    @property
    def stats(self):
        return {}

    @property
    def settings(self):
        return {}


# Attempt to import solvers. These will only be successful if they are built correctly.
try:
    from .cython_glpk import CythonGLPKSolver as cy_CythonGLPKSolver
except ImportError:
    pass
else:
    class CythonGLPKSolver(Solver):
        """Python wrapper of Cython GLPK solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'glpk'

        def __init__(self, *args, **kwargs):
            super(CythonGLPKSolver, self).__init__(*args, **kwargs)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_flows_once', 'PYWR_SOLVER_GLPK_FIXED_FLOWS_ONCE', bool)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_costs_once', 'PYWR_SOLVER_GLPK_FIXED_COSTS_ONCE', bool)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_factors_once', 'PYWR_SOLVER_GLPK_FIXED_FACTORS_ONCE', bool)
            self._cy_solver = cy_CythonGLPKSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            return self._cy_solver.reset()

        def dump_mps(self, filename):
            return self._cy_solver.dump_mps(filename)

        def dump_lp(self, filename):
            return self._cy_solver.dump_lp(filename)

        def dump_glpk(self, filename):
            return self._cy_solver.dump_glpk(filename)

        def retry_solve():
            def fget(self):
                return self._cy_solver.retry_solve

            def fset(self, value):
                self._cy_solver.retry_solve = value

            return locals()
        retry_solve = property(**retry_solve())

        def save_routes_flows():
            def fget(self):
                return self._cy_solver.save_routes_flows

            def fset(self, value):
                self._cy_solver.save_routes_flows = value

            return locals()
        save_routes_flows = property(**save_routes_flows())

        @property
        def routes(self):
            return self._cy_solver.routes

        @property
        def routes_flows_array(self):
            return self._cy_solver.route_flows_arr

        @property
        def stats(self):
            return self._cy_solver.stats

        @property
        def use_presolve(self):
            return self._cy_solver.use_presolve

        @property
        def set_fixed_flows_once(self):
            return self._cy_solver.set_fixed_flows_once

        @property
        def set_fixed_costs_once(self):
            return self._cy_solver.set_fixed_costs_once

        @property
        def set_fixed_factors_once(self):
            return self._cy_solver.set_fixed_factors_once

        @property
        def settings(self):
            return {
                'use_presolve': self.use_presolve,
                'save_routes_flows': self.save_routes_flows,
                'retry_solve': self.retry_solve,
                'set_fixed_flows_once': self.set_fixed_flows_once,
                'set_fixed_costs_once': self.set_fixed_costs_once,
                'set_fixed_factors_once': self.set_fixed_factors_once
            }
    solver_registry.append(CythonGLPKSolver)


try:
    from .cython_glpk import CythonGLPKEdgeSolver as cy_CythonGLPKEdgeSolver
except ImportError:
    pass
else:
    class CythonGLPKEdgeSolver(Solver):
        """Python wrapper of Cython GLPK solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'glpk-edge'

        def __init__(self, *args, **kwargs):
            super(CythonGLPKEdgeSolver, self).__init__(*args, **kwargs)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_flows_once', 'PYWR_SOLVER_GLPK_FIXED_FLOWS_ONCE', bool)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_costs_once', 'PYWR_SOLVER_GLPK_FIXED_COSTS_ONCE', bool)
            kwargs = _parse_env_kwarg(kwargs, 'set_fixed_factors_once', 'PYWR_SOLVER_GLPK_FIXED_FACTORS_ONCE', bool)
            self._cy_solver = cy_CythonGLPKEdgeSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            return self._cy_solver.reset()

        def dump_mps(self, filename):
            return self._cy_solver.dump_mps(filename)

        def dump_lp(self, filename):
            return self._cy_solver.dump_lp(filename)

        def dump_glpk(self, filename):
            return self._cy_solver.dump_glpk(filename)

        def retry_solve():
            def fget(self):
                return self._cy_solver.retry_solve

            def fset(self, value):
                self._cy_solver.retry_solve = value

            return locals()
        retry_solve = property(**retry_solve())

        @property
        def stats(self):
            return self._cy_solver.stats

        @property
        def use_presolve(self):
            return self._cy_solver.use_presolve

        @property
        def set_fixed_flows_once(self):
            return self._cy_solver.set_fixed_flows_once

        @property
        def set_fixed_costs_once(self):
            return self._cy_solver.set_fixed_costs_once

        @property
        def set_fixed_factors_once(self):
            return self._cy_solver.set_fixed_factors_once

        @property
        def settings(self):
            return {
                'use_presolve': self.use_presolve,
                'retry_solve': self.retry_solve,
                'set_fixed_flows_once': self.set_fixed_flows_once,
                'set_fixed_costs_once': self.set_fixed_costs_once,
                'set_fixed_factors_once': self.set_fixed_factors_once
            }
    solver_registry.append(CythonGLPKEdgeSolver)


try:
    from .cython_lpsolve import CythonLPSolveSolver as cy_CythonLPSolveSolver
except ImportError:
    pass
else:
    class CythonLPSolveSolver(Solver):
        """Python wrapper of Cython LPSolve55 solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'lpsolve'

        def __init__(self, *args, **kwargs):
            super(CythonLPSolveSolver, self).__init__(*args, **kwargs)
            self._cy_solver = cy_CythonLPSolveSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            pass

        @property
        def save_routes_flows(self):
            return self._cy_solver.save_routes_flows

        @save_routes_flows.setter
        def save_routes_flows(self, value):
            self._cy_solver.save_routes_flows = value

        @property
        def routes(self):
            return self._cy_solver.routes

        @property
        def routes_flows_array(self):
            return self._cy_solver.route_flows_arr

        @property
        def stats(self):
            return self._cy_solver.stats

        @property
        def settings(self):
            return {
                'save_routes_flows': self.save_routes_flows,
            }
    solver_registry.append(CythonLPSolveSolver)


class NullSolver(Solver):
    """A null solver that performs no allocation routine.

    This solver does no allocation and is mainly useful for debugging purposes.
    """
    name = 'null'

    def setup(self, model):
        pass

    def solve(self, model):
        pass

    def reset(self):
        pass
solver_registry.append(NullSolver)



"""
Tests for `pywr._component.Component` class behaviour and its use in `pywr.core.Model`.
"""

from pywr.core import Component, Model, ModelStructureError
from collections import defaultdict
from fixtures import simple_linear_model
import pytest


class DummyComponent(Component):
    def __init__(self, *args, **kwargs):
        super(DummyComponent, self).__init__(*args, **kwargs)
        self.func_counts = defaultdict(lambda: 0)

    def setup(self):
        self.func_counts['setup'] += 1

    def reset(self):
        self.func_counts['reset'] += 1

    def before(self):
        self.func_counts['before'] += 1
        # Check that all children have been called first.
        for c in self.children:
            assert c.func_counts['before'] == self.func_counts['before']
        # Check that all parents have yet to called.
        for p in self.parents:
            assert p.func_counts['before'] + 1 == self.func_counts['before']


    def after(self):
        self.func_counts['after'] += 1

    def finish(self):
        self.func_counts['finish'] += 1


def test_single_component(simple_linear_model):
    """ Test that a component's methods are called the correct number of times """
    m = simple_linear_model
    nt = len(m.timestepper)
    c = DummyComponent(m)

    m.run()

    assert c.func_counts['setup'] == 1
    assert c.func_counts['reset'] == 1
    assert c.func_counts['before'] == nt
    assert c.func_counts['after'] == nt
    assert c.func_counts['finish'] == 1


def test_shared_parent_component(simple_linear_model):
    """ Test two components sharing the same parent """

    m = simple_linear_model
    nt = len(m.timestepper)
    c1 = DummyComponent(m, name='c1')
    c2 = DummyComponent(m, name='c2')
    cp = DummyComponent(m, name='cp')

    c1.parents.add(cp)
    c2.parents.add(cp)

    assert cp in c1.parents
    assert cp in c2.parents
    assert c1 in cp.children
    assert c2 in cp.children

    m.run()

    assert len(m.components) == 3

    for c in m.components:
        assert c.func_counts['setup'] == 1
        assert c.func_counts['reset'] == 1
        assert c.func_counts['before'] == nt
        assert c.func_counts['after'] == nt
        assert c.func_counts['finish'] == 1

    # Simulate a change and re-run.
    m.setup()
    m.run()

    assert len(m.components) == 3

    for c in m.components:
        assert c.func_counts['setup'] == 2
        assert c.func_counts['reset'] == 3
        assert c.func_counts['before'] == 2*nt
        assert c.func_counts['after'] == 2*nt
        assert c.func_counts['finish'] == 2


def test_circular_components_error(simple_linear_model):
    """ Test that circular components raise an error """
    m = simple_linear_model
    c1 = DummyComponent(m)
    c2 = DummyComponent(m)

    c1.parents.add(c2)
    c2.parents.add(c1)

    with pytest.raises(ModelStructureError):
        m.run()


def test_selfloop_components_error(simple_linear_model):
    """ Test that self-looping components raise an error """
    m = simple_linear_model
    c1 = DummyComponent(m)
    c1.parents.add(c1)

    with pytest.raises(ModelStructureError):
        m.run()
